@@ -86,7 +86,7 @@ static gboolean        toplevel_focus           (BonoboWindow *, GdkEventFocus *
 static void            set_active_view          (BonoboMDI *, GtkWidget *);
 
 /* convenience functions that call child's "virtual" functions */
-static GtkWidget      *child_set_label         (BonoboMDIChild *, GtkWidget *);
+static GtkWidget      *child_set_label         (BonoboMDIChild *, GtkWidget *, GtkWidget *);
 
 static void 	       child_name_changed 	(BonoboMDIChild *mdi_child, 
 						 gchar* old_name, 
@@ -143,6 +143,9 @@ struct _BonoboMDIPrivate
 
 	gint 		 default_window_height;
 	gint		 default_window_width;
+
+	/* Whether our state is being restored for session management */
+	gboolean	 restoring_state;
 };
 
 typedef gboolean   (*BonoboMDISignal1) (GObject *, gpointer, gpointer);
@@ -405,10 +408,10 @@ bonobo_mdi_new (const gchar *mdi_name, const gchar *title,
 }
 
 static GtkWidget *
-child_set_label (BonoboMDIChild *child, GtkWidget *label)
+child_set_label (BonoboMDIChild *child, GtkWidget *view, GtkWidget *label)
 {
 	GtkWidget *w;
-	w = BONOBO_MDI_CHILD_GET_CLASS (child)->set_label (child, label, NULL);
+	w = BONOBO_MDI_CHILD_GET_CLASS (child)->set_label (child, view, label, NULL);
 	
 	return w;
 }
@@ -441,7 +444,14 @@ child_list_activated_cb (BonoboUIComponent *uic, gpointer user_data, const gchar
 		GList *views = bonobo_mdi_child_get_views (child);
 		
 		if (views)
+		{
+			GtkWindow *window;
+
+			window = GTK_WINDOW (bonobo_mdi_get_window_from_view (views->data));
+			gtk_window_present (window);
+
 			bonobo_mdi_set_active_view (mdi, views->data);
+		}
 		else
 			bonobo_mdi_add_view (mdi, child);
 	}
@@ -865,7 +875,7 @@ book_add_view (GtkNotebook *book, GtkWidget *view)
 
 	child = bonobo_mdi_get_child_from_view (view);
 
-	title = child_set_label (child, NULL);
+	title = child_set_label (child, view, NULL);
 
 	gtk_notebook_append_page (book, view, title);
 
@@ -921,6 +931,9 @@ toplevel_focus (BonoboWindow *win, GdkEventFocus *event, BonoboMDI *mdi)
 	/* updates active_view and active_child when a new toplevel receives focus */
 	g_return_val_if_fail (BONOBO_IS_WINDOW (win), FALSE);
 	
+	if (mdi->priv->active_window == win)
+		return FALSE;
+
 	mdi->priv->active_window = win;
 	
 	contents = get_book_from_window (win);
@@ -985,7 +998,7 @@ app_clone (BonoboMDI *mdi, BonoboWindow *win, const char *window_role)
 	
 	app_create (mdi, NULL, window_role);
 
-	if (win != NULL)
+	if (win != NULL && !bonobo_mdi_get_restoring_state (mdi))
 	{
 		const BonoboMDIWindowInfo *window_info = bonobo_mdi_get_window_info (win);
 		g_return_if_fail (window_info != NULL);
@@ -1020,6 +1033,11 @@ app_close_book (BonoboWindow *win, GdkEventAny *event, BonoboMDI *mdi)
 	
 	gedit_debug (DEBUG_MDI, "");
 
+	g_return_val_if_fail (win != NULL, FALSE);
+	g_return_val_if_fail (mdi != NULL, FALSE);
+
+	toplevel_focus (win, NULL, mdi);
+
 	if (g_list_length (mdi->priv->windows) == 1) 
 	{		
 		if (!bonobo_mdi_remove_all (mdi, FALSE))
@@ -1189,9 +1207,10 @@ app_create (BonoboMDI *mdi, gchar *layout_string, const char *window_role)
 	window = bonobo_window_new (mdi->priv->mdi_name, mdi->priv->title);
 	g_return_if_fail (window != NULL);
 
-	gtk_window_set_default_size (GTK_WINDOW (window),
-				     mdi->priv->default_window_width,
-				     mdi->priv->default_window_height);
+	if (!bonobo_mdi_get_restoring_state (mdi))
+		gtk_window_set_default_size (GTK_WINDOW (window),
+					     mdi->priv->default_window_width,
+					     mdi->priv->default_window_height);
 
 	if (window_role)
 		gtk_window_set_role (GTK_WINDOW (window), window_role);
@@ -1318,8 +1337,7 @@ set_active_view (BonoboMDI *mdi, GtkWidget *view)
  * @view: A pointer to the view that is to become the active one.
  * 
  * Description:
- * Sets the active view to @view. It also raises the window containing it
- * and gives it focus.
+ * Sets the active view to @view.
  **/
 void 
 bonobo_mdi_set_active_view (BonoboMDI *mdi, GtkWidget *view)
@@ -1337,8 +1355,6 @@ bonobo_mdi_set_active_view (BonoboMDI *mdi, GtkWidget *view)
 	
 	window = GTK_WINDOW (bonobo_mdi_get_window_from_view (view));
 	
-	gtk_window_present (window);
-
 	set_active_view (mdi, view);
 
 	gedit_debug (DEBUG_MDI, "END");
@@ -1826,7 +1842,7 @@ bonobo_mdi_update_child (BonoboMDI *mdi, BonoboMDIChild *child)
 	{
 		view = GTK_WIDGET (view_node->data);
 
-		title = child_set_label (child, NULL);
+		title = child_set_label (child, view, NULL);
 		
 		gtk_notebook_set_tab_label (GTK_NOTEBOOK (view->parent), view, title);
 		
@@ -2195,3 +2211,46 @@ bonobo_mdi_get_window_info (BonoboWindow *win)
 	return (const BonoboMDIWindowInfo *) 
 			g_object_get_data (G_OBJECT (win), WINDOW_INFO_KEY);
 }
+
+/**
+ * bonobo_mdi_set_restoring_state:
+ * @mdi: A pointer to an MDI object.
+ * @restoring_state: Whether the state is being restored.
+ * 
+ * Sets whether the MDI object is being restored to a known state.
+ **/
+void
+bonobo_mdi_set_restoring_state (BonoboMDI *mdi, gboolean restoring_state)
+{
+	BonoboMDIPrivate *priv;
+
+	g_return_if_fail (mdi != NULL);
+	g_return_if_fail (BONOBO_IS_MDI (mdi));
+
+	priv = mdi->priv;
+
+	g_return_if_fail (priv->restoring_state && restoring_state);
+	g_return_if_fail (!priv->restoring_state && !restoring_state);
+
+	priv->restoring_state = restoring_state ? TRUE : FALSE;
+}
+
+/**
+ * bonobo_mdi_get_restoring_state:
+ * @mdi: A pointer to an MDI object.
+ * 
+ * Queries whether an MDI object is having its state restored.
+ * 
+ * Return value: TRUE if the MDI object is being restored, FALSE otherwise.
+ **/
+gboolean
+bonobo_mdi_get_restoring_state (BonoboMDI *mdi)
+{
+	BonoboMDIPrivate *priv;
+
+	g_return_val_if_fail (mdi != NULL, FALSE);
+	g_return_val_if_fail (BONOBO_IS_MDI (mdi), FALSE);
+
+	priv = mdi->priv;
+	return priv->restoring_state;
+}
